/** sWifiFull: a very easy to use ESP32 WiFi library.
     - Automatic initialization and connection
     - 'stays alive' control
     - eleven freely selectable options
     Totally free to use and modify
     Ewald Kendziorra sr, 2020, www.kendziorra.nl
  ****************************************************/

#define USE_STA        // use WiFi mode STA
#ifdef USE_STA         // only station mode
  #define USE_STAIP    // if using a static STA IP (is needed with: USE_WS)
  #define USE_STADNS   // use WiFi STA DNS
  #define USE_STACHAN  // used in AP_STA mode
  #define USE_OTA      // ArduinoOTA
  #define USE_TIME     // NTP, date, time, timezone, DST/Summertime /** my simpleLIB library: sDateTimeSun.h       **/
//  #define USE_WSER       // Web Server /** library: https://github.com/me-no-dev/ESPAsyncWebServer and https://github.com/me-no-dev/AsyncTCP **/
  #define USE_WS       // WebSockets Server         /**   // used library: https://github.com/me-no-dev/ESPAsyncWebServer and https://github.com/me-no-dev/AsyncTCP **/
//  #define USE_WSCL     // ESP32 Websocket client (sSimpleLib)
  #define USE_WTHR     // Weather  /** my simpleLIB library: sOpenweathermap.h **/
//  #define USE_SOLARMETEO  // KNMI SOLAR weersverwachting /** sSolarMeteo.h **/
#endif
#define USE_AP         // use WiFi mode AP
#ifdef USE_AP          // only softAp mode
  #define USE_APIP     // if using a static AP IP 
  #define USE_APCHAN   // used in AP_STA mode
  #define USE_APHID    // SSID hidden
  #define USE_APMAXCON // max connections, default 4, max 10
  #define USE_NOW      // ESP-NOW only in softAP mode
#endif
#if defined(USE_NOW) || defined(USE_WS)
  #define USE_PARSER   // parse the received message to fieldnames an fieldValues 
#endif
#if defined(USE_STA) && defined(USE_AP) 
  #define USE_LR       // ESP32 softAP Long Range mode up to 1000m (only in AP_STA mode)
#endif

const char staSsid[] = "EDIT_EDIT"@,
            staPassword[] = "EDIT_EDIT"@;
const uint8_t staChannel = 11;  // set this channel FIXED in your ROUTER
const IPAddress staIP(192, 168, 178, 173),
      staGatewayIP(192, 168, 178,   1),
      staSubnetIP(255, 255, 255,   0);
const IPAddress staPrimDNS(8, 8, 8, 8),
      staSecDNS(8, 8, 4, 4);
const char apSsid[] = "apSsid1",
            apPassword[] = "apPassword1";
const uint8_t apChannel = staChannel;
const uint8_t apMaxCon = 6;  //  default 4, max 10
const bool apHiden = true;
const IPAddress apIP(192, 168,   4, 6),
      apGatewayIP(192, 168,   4, 1),
      apSubnetIP(255, 255, 255, 0);

const char myOtaID[] = "sWiFi2_DevC4";  // Hostname OTA (and short board type (DevC4))
const char nowAddress[7] = "000001";  // nodes can have, mixxed, the same or an unique value
// adressing:               "bbffrr"  bb: building, ff: floor, rr: room
//  building: a nowMaxSlaves-value 0-19 (max. 20 locations)
//     floor: a value 00-99 (max. 100 locations)
//      room: a value 00-99 (max. 100 locations)
// floor and room: the nodes themselves test if it is for them
const uint8_t nowMaxSlaves = 2,  // (1-20) the number of buildings in use (it's the second MAC value)
              nowAdrMAC = 6;  // all nodes te same first MAC value
// in ESP-NOW and WebSocket we use this message structure: "|to|from|name|val|name|val|...|...|"
//   ### always a set of two fields ###
const byte nFields = 8,   // the max. number of a set names/values
           nMaxChars = 60;  // the max. length of a message
char fldIn[nFields][nMaxChars + 1],  // parse received message to fieldnames an fieldValues
     valIn[nFields][nMaxChars + 1];  //   ... |toadr|fromadr|field-1|val-1|field-2|val-2|...|...
//        fldIn  valIn   fldIn  valIn  fldIn  valIn ....
const char     wsURI[] = "/sWS";      // WebSocket URI,       also in HTML
const uint16_t wsPort = 82;       // Websocket Port,      also in HTML
char units[] = "metric",   // 'metric' or 'imperial'
       key[] = "EDIT_EDIT"@,  // get a FREE key (60 calls/minute, 1,000,000 calls/month) https://openweathermap.org/price
       lat[] = "52.1234",    // your location
       lon[] = "5.5678";
// ***** List of timezones https://sites.google.com/a/usapiens.com/opnode/time-zones
char timeZone[] = "CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00";  // EU Amsterdam/Barcelona with Summertime (DST)
const time_t syncTime = 12 * 3600;  // NTP time sync interval in sec. if 0 then NOT used
double latSun = atof(lat),
       lonSun = atof(lon);

#include <sWiFiFull.h>  // always under the optional #define xyz declarations 

//##### example code #####//
void nowAction();
void wsAction();
sEpoch ep;     // UNIX epoch ep.sec and  milliseconds in ep.msec
sDateTime dt;  // local/GMT time and date values structure
void timeGet();
void wthrGet();
//########################//

void setup() {
  Serial.begin(115200);
  while (!Serial); delay(1000);
  Serial.println("\n");
  setupWiFi();  // is necessary on this place - init all your defined choices
  //##### example code #####
  Serial.println("simpleWiWi example, choose the options you need!\n"
                 "------------------------------------------------");
  timeGet();  // sDateTimeSun.h, use this in your code where necessary
  wthrGet();  // sOpenweathermap.h, use this in your code where necessary
  //########################//
}

// ============================================= //
// USE MY WebSocket WebMonitorHTML (is included) //
// ============================================= //

void loop() {
  loopWiFi();  // is necessary on this place - connection 'stays alive' control
  //##### example code #####
  if (nowVal) nowAction();  // ESP-NOW data received
  static uint8_t  n = 0;
  static time_t t = millis() + 5000;
  if (millis() > t) {
    char msg[40];
    char toNode[] = "000000";
    sprintf(msg, "|%s|%s|msg|%03u|", toNode, nowAddress, n);
    uint8_t mac = atoi("00");  // first 2 (bb) from to node "bbffrr" address -
    if (nowSendData(mac, msg)) Serial.printf("Success Sent msg: %s\n", msg);
    else Serial.printf("DID NOT SEND msg: %s\n", msg);
    if ((n % 5) == 0) {  // to webbrowser "WS" via address "000000"
      sprintf(msg, "|WS|%s|msg|%03u|", nowAddress, n);
      ws.textAll(msg);  // send to WS
    }
    n++;
    t = millis() + 5000;
  }
  if (wsVal) wsAction();  // WebSocket data received
  static time_t s = 0;
  ep = sNow();
  if (s != ep.sec && (ep.sec % 60) == 0) {  // 60 sec print, sTimestampStr is useful with WebSocket communication
    Serial.printf("The current local time is: %s", ctime(&ep.sec));
    Serial.printf("\tsec: %u, msec: %u, ms-timestamp String: %s\n", ep.sec, ep.msec, sTimestampStr(ep).c_str());
    s = ep.sec;
  }
  yield();
  //########################//
}

//##### example code #####
void nowAction() {
  char msg[nowBuf.length() + 1];
  strcpy(msg, nowBuf.c_str());
  byte len = parseFldVal(msg);
  if (len >= 2) {
    if (strcmp(nowAddress, "000000") == 0) {
      byte toWS = 0;
      for (byte i = 1; i < len; i++) {
        if (strcmp(fldIn[i], "to") == 0 || strcmp(valIn[i], "WS") == 0) toWS = i;
      }
      if (toWS) {
        sprintf(msg, "|WS|%s|", valIn[0]);
        for (byte i = 2; i < len; i++) sprintf(msg + strlen(msg), "%s|%s|", fldIn[i], valIn[i]);
        ws.textAll(msg);  // send to WS
        len = 0;
      }
    }
    // only with node address "000000" or "0000--"
    if (len) {
      char floorAddress[7];
      strncpy(floorAddress, nowAddress, 4);
      sprintf(floorAddress + 4, "--");
      if (len && (strcmp(fldIn[0], nowAddress) == 0 || strcmp(fldIn[0], floorAddress) == 0)) {
        // ##### do something with this data  #####
        Serial.printf("msg from %s:", valIn[0]);
        for (byte i = 1; i < len; i++) Serial.printf(" %s=%s", fldIn[i], valIn[i]);
        Serial.println();
      }
    }
  }
  nowVal = false;
}

void wsAction() {
  char msg[nowBuf.length() + 1];
  strcpy(msg, nowBuf.c_str());
  byte len = parseFldVal(msg);
  if (len >= 2) {
    char adr[8];
    strncpy(adr, nowAddress, 4);
    strcat(adr, "--\0");
    if (strcmp(valIn[0], "WS") == 0) {
      if (strcmp(fldIn[0], nowAddress) == 0 || strcmp(fldIn[0], adr) == 0) {
        // ##### do something with this data  #####
        Serial.printf("msg from %s:", valIn[0]);
        for (byte i = 1; i < len; i++) Serial.printf(" %s=%s", fldIn[i], valIn[i]);
        Serial.println();
        if (strcmp(fldIn[0], nowAddress) == 0) len = 0;
      }
      // only with node address "000000"
      if (strcmp(nowAddress, "000000") == 0) {
        if (len) {
          char c[3] = {fldIn[0][0], fldIn[0][1]};
          uint8_t mac = atoi(c);  // first 2 (bb) from to node "bbffrr" address
          if (nowSendData(mac, nowBuf.c_str())) Serial.printf("Success Sent msg: %s\n", nowBuf.c_str());
          else Serial.printf("DID NOT SEND msg: %s\n", nowBuf.c_str());
        }
      }
    }
  }
  wsVal = false;
}

void timeGet() {
  //  time: print as examples
  bool gmt = false;  // true: UTC/GMT values
  //sRiseSet rs = sSun( now(), lat, lon, gmt);
  sRiseSet rs = sSun( now(), latSun, lonSun);
  dt = sTimeDate(gmt);
  Serial.printf("SunRise: %02u:%02u, SunSet: %02u:%02u\n", hour(rs.rise), minute(rs.rise), hour(rs.set), minute(rs.set));
  time_t t = now();  // or use now(value): value is a + or - offset to the actual local time
  Serial.printf("sDateTimeSun.h - time_t t = now(),\n"
                "day(t), month(t), year(t),\n"
                "hour(t), minute(t), second(t),\n"
                "weekday(t), yweek(t), yday(t):\n\t"
                "%02u-%02u-%04u %02u:%02u:%02u local time "
                "weekday: %02u, yweek: %u, yday: %03u\n",
                day(t), month(t), year(t),
                hour(t), minute(t), second(t),
                weekday(t), yweek(t), yday(t));  // without the value (t) in the functions.... the time goes on
  // on a case-by-case basis you can also use these functions without value (t)
  //  dt-structure: print as example
  Serial.printf("sDateTimeSun.h dt-structure, time with milliseconds:\n\t"
                "%02d:%02d:%02d.%03d %s %02d %s (%02d) %02d  "
                "wday:%2d, week:%3d, yday:%4d, ",
                dt.hour, dt.minute, dt.second, dt.msec,         // time with milliseconds
                dt.dayS, dt.day, dt.monthS, dt.month, dt.year,  // month, day and year
                dt.wday, dt.week, dt.yday);                     // diverse
  if (!gmt) Serial.printf("DST/Summertime: %s\n", (dt.dst > 0 ? "Y" : "N"));  // DST
  else Serial.println("GMT");
}

void wthrGet() {
  getWeather();
  //  owm-structure: print as example
  Serial.println("sOpenweathermap.h owm-structure:");
  Serial.printf("\ttimeZone\t%s sec\n", owm.zone);
  Serial.printf("\tsunRise \t%s:%s:%s UTC\n", owm.rise[0], owm.rise[1], owm.rise[2]);
  Serial.printf("\tsunSet  \t%s:%s:%s UTC\n", owm.set[0], owm.set[1], owm.set[2]);
  Serial.printf("\ttemperature\t%s*C\n", owm.temp);
  Serial.printf("\thumidity\t%s%%\n", owm.hum);
  Serial.printf("\tpressure\t%s hPa\n", owm.pres);
  Serial.printf("\twindSpeed\t%s m/s\n", owm.wind);
  Serial.printf("\twindDir.\t%s\n", owm.dir);
}
//########################//
