#ifndef SIMPLEDB_SD_H
#define SIMPLEDB_SD_H

#include <SPI.h>
#include <SD.h>

enum DB_RESULT {  // result code
 DB_OK       =  0,  // OK, no error
 DB_EMPTY    =  1,  // no record(s) in db
 NOT_FOUND   =  2,  // record not found
 REC_EXIST   =  3,  // record already exists
 /** ERRORs **/
 DIF_NR      =  4,  // difference in number of records (DB and backup DB)
 CORRUPT     =  5,  // db corrupt
 MAX_CHARS   =  6,  // key or data: too many characters
 MAX_REC     =  7,  // max records in db
 SEEK_ERR    =  8,  // file seek error  
 EXIST_ERR   =  9,  // no file
 BACKUP_ERR  = 10,  // backup problem
 FORMAT_ERR  = 11,  // format problem
 TOMBSTONE   = 12,  // deleted record
 REMOVE_ERR  = 13,  // delete file error
 RENAME_ERR  = 14,  // rename file error
 MOUNT_ERR   = 15,  // can not mount  
 OPEN_ERR    = 16,  // file open error  
 DB_ERR      = 17,  // db error
};

uint8_t tsM,         // pos. tombstone marker
        nxW;         // fixed index width (key width + tombstone marker "key|0|")
char    hlK[7],      // strcpy(), holder key format string
        hlR[7],      // strcpy(), holder record format string
        *nxB;        // index read buffer
uint8_t  dtW;        // max data width in a record, exclusive 'key' and 'tombstone' marker
uint16_t dbR,        // record counter, exlusive tombstone records
         dbT;        // record counter, inclusive tombstone records
char     *dbB;       // record read buffer
DB_RESULT dbResult;  // DB_RESULT error code holder
File dbF, cpF;  // DB and DB-backup 

/** init an new or existing database **/
DB_RESULT initDB();  // SD
/** simpleDB_SD functions **/
DB_RESULT nrRec();                                      // set counter dbR, the nummer of records, tombstone records are excluded
DB_RESULT addRec(const char *key, const char *data);    // add a record with a UNIQUE key 
DB_RESULT writeRec(const char *key, const char *data);  // save a record, overwrite if key exist else add a record 
DB_RESULT readRec(const char *key);                     // read the record with the given "key", it is in record buffer dbB 
DB_RESULT getRec(uint8_t nr);                           // read a record on position "nr", it is in record buffer dbB 
DB_RESULT findRec(const char *key);                     // find record with the given "key", slow but simple and fast enough for a small db 
DB_RESULT delRec(const char *key);                      // tombstone a record, remove tombstones with comprDb 
DB_RESULT cmprDb();                                     // copy db, then remove tombstone records 
DB_RESULT openDB();                                     // test if open, else open or create DB
void      closeDB();                                    // close db
DB_RESULT wrtRec(const char *key, const char *data, bool eof);  // PRIVATE function: over write or add a record 

/** init an new or existing SD datanase **/
DB_RESULT initDB() {
  // init DB parameters
  tsM = db_KeyLen + 1;     // pos. tombstone marker
  nxW = db_KeyLen + 3;     // index width (key width + tombstone marker "key|0|")
  dtW = db_RecLen - nxW;   // max data width in a record, exclusive 'key' and 'tombstone' marker
  dbB = new char[db_RecLen],      // record read buffer
  nxB = new char[nxW + 1];  // index read buffer

  // used in addRec(): create a strcpy() format string (fixed record size)
  strcpy(hlK,("%" + String(db_KeyLen) + "s|0|").c_str());  // if kwW =   4 then hlK = "%4s|0|",  "|" is delimiter, "0" if tombstone
  strcpy(hlR,("%-" + String(dtW) + "s").c_str());    // if dtW = 104 then hlR = "%-104s"
  if (!SD.begin(db_SsPin)) {dbResult = MOUNT_ERR; return dbResult;}
  openDB();  // SD start,
  return dbResult;
}

/** open db **/
DB_RESULT openDB() {
  dbResult = DB_OK;
  if (dbF && (dbR == 0)) dbResult = nrRec();
  else {
   dbF = SD.open(db_Name, FILE_WRITE);
   if (!dbF) dbResult = OPEN_ERR;
   else if (dbR == 0) dbResult = nrRec();
  }
  return dbResult;
}

/** close db **/
void closeDB() {
  dbF.close();
}

/** set counter dbR, the nummer of records, tombstone records are excluded **/
DB_RESULT nrRec() {  // initDB() call this function, simpleDB_SD update dbR it and you can use dbR
  dbResult = DB_OK;
  uint16_t len;
  dbR = dbT = 0;
  if (dbF.size() < db_RecLen) return dbResult;  // db is empty
  if (!dbF.seek(0)) return dbResult;
  while (dbF.available()) {
   len = dbF.readBytes(nxB, nxW);
   nxB[len] = '\0';
   if (len != nxW) {dbResult = CORRUPT; break;}
   if (nxB[tsM] == '0') dbR++; // it's not a tombstone
   dbT++;
   if (!dbF.seek(dbF.position() + dtW)) break;  // skip data, set entree to next record
  }
  return dbResult;
}

/** add a record with a UNIQUE key **/
DB_RESULT addRec(const char *key, const char *data) {  // key UNIQUE!!
  if (openDB() != DB_OK) return dbResult;
  if (dbR == db_RecLen) {dbResult = MAX_REC; return dbResult;}  // max records in db
  if (strlen(data) > dtW) {dbResult = MAX_CHARS; return dbResult;}  // data: too many characters
  if (findRec(key) == REC_EXIST) return REC_EXIST;  // record already exists
  wrtRec(key, data, true);
  return dbResult;
}

/** save a record, overwrite if key exist else add a record **/
DB_RESULT writeRec(const char *key, const char *data) {
  if (openDB() != DB_OK) return dbResult;
  if (strlen(data) > dtW) {dbResult = MAX_CHARS; return dbResult;}  // data: too many characters
  if (findRec(key) > REC_EXIST) return dbResult;  // is there a find error?
  wrtRec(key, data, (dbResult ==  REC_EXIST ? false : true));
  return dbResult;
}

/** PRIVATE function: over write or add a record **/
DB_RESULT wrtRec(const char *key, const char *data, bool eof) {
  dbResult = DB_OK;
  if (eof) if (!dbF.seek(dbF.size())) {dbResult = SEEK_ERR; return dbResult;}  // if record not exists then seek to EOF
  // format key and tombstone marker "key|0|" 
  char n[db_RecLen + 1];
  sprintf(n, hlK, key);
  dbF.print(n);  // write key + tombstone to db
  // format data with trailing white spaces
  sprintf(n, hlR, data);
  dbF.print(n);    // write fields to db
  dbF.flush();     // save record physically to the SD
  if (eof) {dbR++; dbT++;}  // update record nr
  return dbResult;
}

/** read the record with the given "key", it is in record buffer dbB **/
DB_RESULT readRec(const char *key) {
  if (openDB() != DB_OK) return dbResult;
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  if (findRec(key) != REC_EXIST) return dbResult;  // find position in file
  if (!dbF.seek(dbF.position() + nxW)) {dbResult = SEEK_ERR; return dbResult;}  // seek to first position of data (the '|' after tombstone marker)
  uint8_t len = dbF.readBytes(dbB, dtW);  //  get only the data, data in buffer dbB
  dbB[len] = '\0';
  if (len != dtW) dbResult = CORRUPT;
  return dbResult;
}

/** read a record on position "nr", it is in record buffer dbB, included tombstone records **/
DB_RESULT getRec(uint8_t nr) {  // nr: position of record
  if (openDB() != DB_OK) return dbResult;
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  if (dbF.size() < db_RecLen) {dbResult = DB_EMPTY; return dbResult;}  // db is empty
  if (!dbF.seek(nr * db_RecLen)) {dbB[0] = '\0'; dbResult = SEEK_ERR; return dbResult;}
  uint8_t len = dbF.readBytes(dbB, db_RecLen);  // record in buffer dbB
  dbB[len] = '\0';
  if (dbB[tsM] != '0') dbResult = TOMBSTONE; // if it is a tombstone
  if (len != db_RecLen) dbResult = CORRUPT;
  return dbResult;
}

/** find record with the given "key", slow but simple and fast enough for a small db **/
DB_RESULT findRec(const char *key) {
  // seek to record, sequentieel search
  // false: not found, true: found
  if (strlen(key) != db_KeyLen) {dbResult = MAX_CHARS; return dbResult;}
  if (openDB() != DB_OK) return dbResult;
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  if (!dbF.seek(0)) {dbResult = SEEK_ERR; return dbResult;}  
  if (dbF.size() < db_RecLen) {dbResult = DB_EMPTY; return dbResult;}  // db is empty
  uint16_t len;
  while (dbF.available()) {
   len = dbF.readBytes(nxB, nxW);
   nxB[len] = '\0';
   if (len != nxW) {dbResult = CORRUPT; break;}
   if (nxB[tsM] == '0') {  // not a tombstone?
     bool f = true;
     for (uint8_t i = 0; i < db_KeyLen; i++) {
       if (nxB[i] != key[i]) {f = false; break;}
     }
     if (f) {
       if (!dbF.seek(dbF.position() - nxW)) break;  // seek to the start byte of the record
       dbResult = REC_EXIST;
       return dbResult;  // key is found
     }
   }
   if (!dbF.seek(dbF.position() + dtW)) break;  // skip data, set entree to next record
  }
  return NOT_FOUND;
}

/** tombstone a record, remove tombstones with comprDb **/
DB_RESULT delRec(const char *key) {
  if (openDB() != DB_OK) return dbResult;
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  if (findRec(key) != REC_EXIST) return dbResult;  // find position in file
  if (!dbF.seek(dbF.position() + tsM)) {dbResult = SEEK_ERR; return dbResult;}
  dbF.print('1');
  dbF.flush();
  dbR--;  // update record nr
  return dbResult;
}

/** copy db, then remove tombstone records **/
DB_RESULT cmprDb() { 
  if (dbR == 0) {dbResult = DB_EMPTY; return dbResult;}
  dbResult = DB_OK;
  dbF.close();
  if (openDB() != DB_OK) return dbResult;
  if (SD.exists(db_Backup)) if (!SD.remove(db_Backup)) return REMOVE_ERR;  // delete file error
  cpF = SD.open(db_Backup, FILE_WRITE);
  if (!cpF) {dbResult = OPEN_ERR; return dbResult;}
  cpF = SD.open(db_Backup, FILE_WRITE);
  if (!cpF) {dbResult = OPEN_ERR; return dbResult;}
  // make a copy
  if (!dbF.seek(0)) {dbResult = SEEK_ERR; return dbResult;}
  while (dbF.available()) cpF.print((char)dbF.read());
  cpF.flush();
  dbF.close();
  // delete db
  if (!SD.remove(db_Name)) return REMOVE_ERR;  // delete file error

  // make a new db file
  if (openDB() != DB_OK) return dbResult;
  if (dbF.size() != 0) return EXIST_ERR;  // file not earased - is not possible?

  // copy the 'not as tombstone marked' records from copy to database
  uint16_t len;
  int16_t nr = 0;
  if (!cpF.seek(0)) {dbResult = SEEK_ERR; return dbResult;}
  while (cpF.available()) {
   len = cpF.readBytes(dbB, db_RecLen);
   dbB[len] = '\0';
   if (strlen(dbB) != len) {
     cpF.close();
     dbResult = CORRUPT;
     break;  // db corrupt
   }
   if (dbB[tsM] == '0') {  // if it is not a tombstone ...
     dbF.print(dbB);  // ... then copy it to the tmp file
     dbF.flush();
     nr++;
   }
  }
  cpF.close();  // do not remove, it is your backup
  dbR = nr;
  if (nr != dbR) return DIF_NR;  // difference in number of records
  return dbResult;
}

#endif
