  // a very simple ESP32 local and GMT time solution, totally free to use and modify
  // Ewald Kendziorra sr, 2023, http://www.kendziorra.nl 

#ifndef _SIMPLE_TIME_H
  #define _SIMPLE_TIME_H

  #include "time.h"  // (ESP32)
  
  struct sRiseSet {    // output: SunRise and SunSet local or UTC
    time_t rise,       // SunRise local or UTC
           set;        // SunSet  local or UTC
  };
  bool getNTP();  // get UNIX timestamp van time server(s)
  void setDayTime(time_t t);  // set time manual, use sMakeTime() to get t
  bool timeOk();                      // false: time NOT OK, use function getNTP()
  /** USER TIME FUNCTIONS **/         // now(): get the local UNIX timestamp of this day
  time_t      now(time_t t = 0);      // now(+/- val): another date/time, add/subst n-seconds
  uint16_t   year(time_t t = now());  // get the NOW date year(), ....., second() values  
  uint8_t   month(time_t t = now());  //   or add/subst n-seconds for a different time or date 
  uint8_t     day(time_t t = now());  //     suppose it is now 12h then: 
  uint16_t   yday(time_t t = now());  //       hour() = 12h, and hour(3600 * 5) = 17h
  uint8_t    hour(time_t t = now());
  uint8_t  minute(time_t t = now());
  uint8_t  second(time_t t = now());
  time_t   getGmt(time_t t = now());  // from Local to GMT time
  time_t makeTime(int8_t h, int8_t m, int8_t s, int8_t D, int8_t M, int16_t Y);  // calc a time
  time_t getZone(time_t t = now());   // timezone offset
  time_t getDst(time_t t = now());    // DST/Summertime offset (DST - Local)
  time_t getLocOffset(time_t = now());// total offset GMT to Local
  sRiseSet sSun(time_t t, double lat, double lon, bool gmt = false);  // calc rise and set times
  bool sDaylight(time_t t, sRiseSet rs);  // is it sun down time 
  /********** END **********/
  time_t tmOk = 946681200; // >= Jan 1 2000 00:00:00
  const char * NTPservers[21] = {"time.google.com", "pool.ntp.org", "time.nist.gov"};  // standard NTP Servers
  int16_t getTval(time_t t, char cc, bool gmt = false);  // private ;=)
  time_t sCalcSun(double lat, double lon, uint16_t N, bool rise);    // calc sun time

  timeval ts;  // ts.tv_sec: UNIX timestamp, ts.tv_usec

  bool getNTP() {
    if (WiFi.isConnected()) {
      configTzTime(timeZone, (char*)NTPservers[0], (char*)NTPservers[1], (char*)NTPservers[2]);
      gettimeofday(&ts, nullptr);
      time_t i = ts.tv_sec;
      while(ts.tv_sec < tmOk) {  // time OK?
        gettimeofday(&ts, nullptr);
        if ((ts.tv_sec - i) > 30000) break;  // after 30 sec.
      }
    }
    return ts.tv_sec > tmOk;  // if false: try again later
  }

  void setDayTime(time_t t) {  // set time manual, use sMakeTime()
    timeval ts = {ts.tv_sec = t};
    settimeofday(&ts, nullptr);
  }

  bool timeOk() {
    return now() > tmOk;
  }
  
  time_t now(time_t t) {
    time_t t1;
    gettimeofday(&ts, nullptr);
    return ts.tv_sec + t;
  }

  uint16_t year(time_t t) {
    return getTval(t, 'Y');
  }

  uint8_t month(time_t t) {
    return getTval(t, 'm');  
  }

  uint8_t day(time_t t) {
    return getTval(t, 'd');   
  }

  uint16_t yday(time_t t) {
    return getTval(t, 'j');   
  }

  uint8_t weekday(time_t t) {
    return getTval(t, 'u');   // 'u': 1=monday, 7=sunday - 'w':0=sunday,6=saturday 
  }
  
  uint8_t hour(time_t t) {
    return getTval(t, 'H');   
  }

  uint8_t minute(time_t t) {
    return getTval(t, 'M');   
  }

  uint8_t second(time_t t) {
    return getTval(t, 'S');   
  }
  
  time_t getGmt(time_t t) {
    tm *tmp = gmtime(&t);
    return mktime(tmp);
  }
  
  int16_t getTval(time_t t, char cc, bool gmt) {
    char b[8], c[3] = "%";
    tm * tmp = (gmt ? gmtime(&t) : localtime(&t)); //
    c[1] = cc;
    strftime(b, 5, c, tmp);
    return atoi(b);
  }

  time_t makeTime(int8_t h, int8_t m, int8_t s, int8_t D, int8_t M, int16_t Y) {
    tm tmp;
    tmp.tm_sec = s;
    tmp.tm_min = m;
    tmp.tm_hour = h;
    tmp.tm_mday = D;
    tmp.tm_mon = M - 1;
    tmp.tm_year = Y - 1900;
    time_t mkt = mktime(&tmp);
    return mkt;
  }

  time_t getZone(time_t t) {
    time_t gmt;
    tm *ptm;
    ptm = gmtime(&t);
    gmt = mktime(ptm);
    return (time_t)difftime(t, gmt);
  }

  time_t getDst(time_t t) {
    time_t gmt;
    tm *ptm;
    ptm = gmtime(&t);
    ptm->tm_isdst = -1;
    gmt = mktime(ptm);
    return difftime(t, gmt) - getZone(t);
  }

  time_t getLocOffset(time_t t) {
    return getZone(t) + getDst(t);
  }
  
  /** calc sunrise and sunset times **/
  sRiseSet sSun(time_t t, double lat, double lon, bool gmt) {  // calc rise and set times
    sRiseSet sun;    
    sun.rise = sCalcSun(lat, lon, yday(t), true);  // get SunRise time 
    sun.set  = sCalcSun(lat, lon, yday(t), false);  // get SunSet time
    if (!gmt) {
      time_t v = getLocOffset(t);
      sun.rise += v;
      sun.set += v;
    }
    return sun;  // result: _sun.rise and _sun.set
  }

  bool sDaylight(time_t t, sRiseSet rs) {  // is it sun down time 
    time_t m[3] = {hour(t) * 60 + minute(t), 
                     hour(rs.rise) * 60 + minute(rs.rise),
                     hour(rs.set) * 60 + minute(rs.set)};
    return m[0] > m[1] && m[0] < m[2];
  }

  /** Source:
   *  Almanac for Computers, 1990
   *  published by Nautical Almanac Office
   *  United States Naval Observatory
   *  Washington, DC 20392
   *  Adapted by Ewald Kendziorra sr, www.kendziorra.nl
   **/
  time_t sCalcSun(double lat, double lon, uint16_t N, bool rise) {
    const double _zenith = 90.883,
                 _toRad = 0.01745329,
                 _toDeg = 57.2957786;
    double lngHour = lon / 15.0;
    double t =  N + ((rise ? 6.0 : 18.0) - lngHour) / 24.0;
    double M = (0.9856 * t) - 3.289;
    double L = M + (1.916 * sin(M * _toRad)) + (0.020 * sin(2 * M * _toRad)) + 282.634;
    L += (L < 0 ? 360 : L > 360 ? -360 : 0); 
    double RA = atan(0.91764 * tan(L * _toRad)) * _toDeg;
    RA += (RA < 0 ? 360 : RA > 360 ? -360 : 0);
    double Lquadrant = (floor(L / 90.0)) * 90.0;
    double RAquadrant = (floor(RA / 90.0)) * 90.0;
    RA = (RA + Lquadrant - RAquadrant) / 15.0;
    double sinDec = 0.39782 * sin(L * _toRad);
    double cosDec = cos(asin(sinDec));
    double cosH = (cos(_zenith * _toRad) - (sinDec * sin(lat * _toRad))) / (cosDec * cos(lat * _toRad));
    double H = (rise ? (360.0 - acos(cosH) * _toDeg) : (acos(cosH) * _toDeg)) / 15; 
    double T = H + RA - (0.06571 * t) - 6.622;
    double UT = T - lngHour;
    UT +=  (UT < 0 ? 24 : UT > 24 ? -24 : 0);
    //return UT;  // UTC decimal time 
    UT -= 1.0;
    uint8_t hh = (uint8_t)UT;
    uint8_t mm = (uint8_t)((UT - hh) * 60);
    return hh * 3600 + mm * 60;
  }

#endif  
